{
/////////////////////////////////////////////////////////////////////////////

FILENAME: C3_SD_DRV_010.SPIN

AUTHOR: Andre' LaMothe, Nurve Networks LLC
LAST MODIFIED: 9.3.10
LICENSE: MIT

I. DESCRIPTION

This driver contains 3 layers of software to faciliate the interfacing of SD cards,
specifically the C3 SD Storage Card. However, the functions can be used with
any propeller chip interface simply by changing the SPI bus pins and modifying the CS "chip select"
line logic since the C3 uses a counter and decoder to select SPI CS lines, you must modify the function:

   SPI_Select_Channel( channel# )

   To simply assert the CS line for your particular hardware (if not the C3). Likewise,
   if you want to use another SD card driver for the C3, then all you need to do is find
   function (code) that asserts and de-asserts the CS and replace it with the C3 compatible SPI
   bus function SPI_Select_Channel( channel# ), or similar code. 

The 3 layers to communicate to the SD card are:

----------------------------------------------------------------------------------------
|     FAT16 (High Level - read DOS formatted FAT16 file system files on card)          |
----------------------------------------------------------------------------------------
|     SD Card Interface (Medium Level - initialize card, read, write sectors, etc.)    |
----------------------------------------------------------------------------------------
|     SPI Interface (Low Level - send and receive bytes directly to the SPI interface) |
---------------------------------------------------------------------------------------- 

And of course sitting on top of the FAT16 layer is the application itself.

The driver consists of function calls to access all 3 layers of the model, so you can do
things like access the SD card as if it were a disk drive, print the directory, load files
of any size, etc. Or you might want to forget using DOS FAT16 format, and just read and write
sectors directly to the SD card, lastly, you might want to modify the SPI routines and code
them in assembly language.

Everything is written in SPIN for ease of use and review to see what's going on and the functions
are heavily commented.

SPI Interface Links:

  http://www.vti.fi/midcom-serveattachmentguid-b469d17c67e60d4e3dbb25b0d099ad68/TN15_SPI_Interface_Specification.pdf

SD Card Links:

  http://www.cs.ucr.edu/~amitra/sdcard/ProdManualSDCardv1.9.pdf
  http://www.sandisk.com/Assets/File/OEM/Manuals/SD_SDIO_specsv1.pdf

FAT16 Links:

  http://www.maverick-os.dk/FileSystemFormats/FAT16_FileSystem.html
  http://www.compuphase.com/mbr_fat.htm


II. API OVERVIEW

The following is a short list of the API functions with single line explanations. Refer to the CON and VAR sections
below for insight into the data structures used by the API.


*** Initialization Functions ***

  PUB Start ( buffer_ptrs ) - Initializes the driver and writes the callers pointer storage space with local disk buffer addresses 

*** SPI Functions ***

  PUB SPI_Init(mode) - Initializes the hardware SPI interface I/O pins from propeller to card. 

  PUB SPI_Send_Recv_Byte(spi_data8) - Sends a single byte to SPI interface and receives one at the same time. 

  PUB SPI_Read_Byte - Reads the SPI buffer, writes a dummy values of $FF. 

  PUB SPI_Write_Byte(spi_data8) - Writes a byte to the SPI interface. 


*** SD Functions ***

  PUB SD_Mount - Mounts the SD card by initializing and placing it into SPI mode.

  PUB SD_Unmount - Unmounts the previously mounted SD card.

  PUB SD_Send_Command( command8, address32 ) - Sends a generic command to the SD card. 

  PUB SD_Read_Sector(sector32, sectorbuffer16) - Reads a sector from the SD card.

  PUB SD_Write_Sector(sector32, sectorbuffer16) - Writes a sector to the SD card.

  PUB SD_Wait_Write_Complete - Internal function used in the write command to make sure the flash has been updated.

  PUB SD_Print_Sector(sector32, sect_ptr, start_byte, end_byte, base, print_addr) - Diagnostic function that prints the contents of a sector to terminal.

  PUB SD_Read_WP - reads the single bit "write protect" signal on the SD card mechanical (not implemented on C3)

  PUB SD_Read_CD - reads the single bit "card inserted" signal on the SD card mechanical (not implemented on C3)

*** FAT16 Functions ***

Lower level functions (initialization of FAT16 system)

  PUB FAT_Read_MBR(mbr_ptr) - Reads the MBR (Master Boot Record) from the SD card, sector 0.

  PUB FAT_Load_Partition_Entry(partition, mbr_ptr) - Loads the requested 16-byte partition entry from the loaded MBR.

  PUB FAT_Print_Partition_Entry(partition, mbr_ptr) - Pretty prints the partition entry to terminal.

  PUB FAT_Load_Partition_Boot_Rec(pbr_ptr) - Loads the partition boot record referred to by loaded partition entry.

  PUB FAT_Print_Partition_Boot_Rec - Pretty prints the partition boot record (aka volume record or simply boot record).

High level functions (general file I/O)

  PUB FAT_Print_Directory - Print the root file directory to the terminal (like DOS DIR command).
  
  PUB FAT_File_Open(filename_ptr, file_handle_ptr) - Opens filename and fills in the sent file handle structure.

  PUB FAT_File_Close(file_handle_ptr) - Closes the file handle.

  PUB FAT_File_Read(file_handle_ptr, buffer_ptr, count) - Reads bytes from file referred to by the sent file handle.

  PUB FAT_File_Seek(file_handle_ptr, count, mode) - Seeks file pointer to a specific location in file for reading.

*** Utility Functions ***

  PUB To_ASCII(inchar, replace_char) - Used to map non-printable characters to printable. 

  PUB itoa(value, sptr) - Converts an integer to an NULL terminated ASCII string.

  PUB ToUpper(char) - Converts lower case ASCII to upper case. 

  PUB Strncomp(string_ptr1, string_ptr2, length) - Compares strings.

  PUB _Min(a,b) - returns the min.

  PUB _Max(a,b) - returns the max.

  PUB Set_Debug_Level ( level ) - sets the debug level

*** Exported Object Functions ***

  PUB lcd_bin( value, digits ) - Prints a binary number to the terminal with specific number of digits.
  
  PUB lcd_out( char ) - Outputs a single character to terminal.
       
  PUB lcd_dec( value ) - Prints a number in decimal format to terminal.
 
  PUB lcd_hex( value, digits ) - Prints a hex number to the terminal with specific number of digits.
  
  PUB lcd_setx( new_x ) - Sets the x cursor position on terminal.
 
  PUB lcd_sety( new_y ) - Sets the y cursor position on terminal.
  
  PUB lcd_getx - Gets the current x cursor position in terminal.
 
  PUB lcd_gety - Gets the current y cursor position in terminal.

  PUB lcd_pstring( string_ptr ) - Prints a NULL terminated string to the terminal.
  
  PUB lcd_erase - Erases completely the character under the current cursor position.

  PUB lcd_clear Erases the complete display set cursor to x = 1 ,y = 1

  PUB lcd_crlf  carriage return and linefeed

  PUB lcd_cr    carriage return

  PUB lcd_lf    linefeed

III. NOTES

1. Make sure to refer to the return value from each function, the functions have non-homgeneous
   return codes. For example, in some cases 0 is an error, in some cases it means success.

2. Many of the functions have sprinkled terminal print statements that are left in them
   incase you want to output more debug information. If you don't want any NTSC screen printing
   support simply remove all references to term.* calls in the code and don't include the terminal
   driver.

3. The functions are written for maximum clarity, not speed. I suggest re-writing everything
   in ASM, but these SPIN versions are reasonably easy to follow, so you can learn the ropes
   of SPI, SD, FAT16 drivers.

4. The FAT16 driver layer is minimal and doesn't write files, and only supports the root
   directory and no folders. But, that should suffice for 99% of applications.

5. Everything has been tested, but don't assume everything is working perfectly, if
   something doesn't work as expected test it thouroughly, but it could very well the
   drivers. They were only tested funtionally, but not in any large application, so
   they are not threshed out and probably have bugs here and there. But, nonetheless
   good for tutorial use and to get you started, so you dont' have to write your own!

7. Remember, ALWAYS format your SD cards FAT16, and make sure you do NOT use
   folders and all your file names are ALWAYS 8.3. Also, its suggested that you
   use 1-2G SD cards, these seem to work the best. 

8. Note that this program uses TV_TERMINAL_011.DRV, this is slightly modified from the original
   Parallax source in that some functions to set/get the cursor position have been added
   to make printing easier. Also, notice that the terminal functions have been exported to the
   caller thru an "interface", so the caller can print to the terminal as well and not load
   another tv terminal object.

}


CON

   ' SPI (serial peripheral interface) interface pins
   spiDO  = 10
   spiCLK = 11
   spiDI  = 9

   'spiCS  = 19 ' not used on C3, instead SPI_Select_Channel( channel# ) is used

   ' used on Gam_Bo_Prop to expect a SD-modulor sD-card
   WP   = 6
   CD   = 7

  
   ' SPI pins
   'SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
   'SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW
   SPI_SEL0               = 0
   SPI_SEL1               = 1
   SPI_SEL2               = 2
   SPI_MOSI               = 9  ' SPI master out serial in to slave (spiDI)
   SPI_MISO               = 10  ' SPI master in serial out from slave (spiDO), 10 for good C3, change do 7 for BROKEN C3 with YELLOW jumper.
   SPI_SCK                = 11 ' SPI clock from master to all slaves
  

   ' SD card (secure digital) constants (you can reduce these to speed things up and see if the SD card still works)
   SD_RESPONSE_LIMIT        = 256     ' maximum number of iterations that SD card will be querried for response after a command sent
   SD_POWERUP_LIMIT         = 256     ' maximum number of iterations to initially clock SD card to power up
   SD_RETRIEVE_SECTOR_LIMIT = 4096    ' maximum number of iterations for SD to retrieve a requested sector/block
   SD_WRITE_SECTOR_LIMIT    = 128     ' maximum number of iterations for SD to write a requested sector/block

   ' SD commands (based on SD spec names)
   SD_GO_IDLE_STATE     = $40         ' resets the SD card, places it into SPI mode, idle                              
   SD_EXIT_IDLE_STATE   = $41         ' takes SD out of idle state ready for commands

   SD_APP_CMD           = ($40 + $37) ' makes sure card is in SD mode
   SD_SEND_OP_CMD       = ($40 + $29) ' makes sure card is in SD mode
   SD_READ_SINGLE_BLOCK = ($40 + $11) ' reads a single block (usually a 512 byte sector) at byte address sent in command address
   SD_WRITE_BLOCK       = ($40 + $18) ' write a single block (usually a 512 byte sector) at byte address sent in command address   

   SECTOR_SIZE          = 512         ' should always be 512 for SD cards                  
   MAX_SECTORS_BUFFER   = 1           ' number of sectors buffer can hold, make larger if needed 

   ' fat constants for various functions
   FILE_SEEK_ABS        = 0           ' interpret seek request with absolute position                                         
   FILE_SEEK_REL        = 1           ' interpret seek request as relative position added to current position
   FILE_SEEK_END        = 2           ' seek to end of file   (file_size - 1)
   FILE_SEEK_START      = 3           ' seek to start of file (0)

   ' debugging constants for function terminal debug outputs
   DEBUG_OFF            = 2           ' no output at all
   DEBUG_VERBOSE        = 1           ' verbose mode for functions debug output
   DEBUG_BRIEF          = 0           ' brief mode for functions debug output
   

VAR

  ' API GLOBALS----------------------------------------------------------------

  ' debug values
  long debug_level              ' setting of printout for debug level (DEBUG_OFF, DEBUG_VERBOSE, DEBUG_BRIEF)
  
  long debug_wait_res_num_iter  ' the number of times the sd send command function internally waits for a SPI response

  ' these are the local working buffers to hold the master boot record, partition boot record, fat sectors, and general diskbuffer to
  ' read sectors, in general the driver needs and uses all of these, but to save memory, so the caller doesn't have to instantiate his
  ' own buffers, pointers to these are passed BACK to the caller when he calls the Start method, then he can use the storage buffers
  ' and access, inspect them as well, currently 2K bytes is used by all buffers. Technically, once the MBR and PBR have been loaded
  ' they can be destroyed and the memory used, but then you would have to reload them to review or inspect, thus, suggest leave them
  ' alone after loading...fatbuff is used internally by the file functions, thus between calls, diskbuff is not used by any of the calls, so
  ' will remain intact call to call. However, when debug_level is verbose then diskbuff is used in a couple function to load sector data to
  ' print therefore,it will NOT remain intact call to call, but in debug level brief and off, diskbuff is NOT used by the driver internally
  ' and can be used by the caller for storage during the file read operations
   
  byte diskbuff[ SECTOR_SIZE*MAX_SECTORS_BUFFER ]  ' generic disk sector buffer
  byte mbrbuff[ SECTOR_SIZE ] ' master boot record buffer
  byte pbrbuff[ SECTOR_SIZE ] ' partition boot record
  byte fatbuff[ SECTOR_SIZE]  ' holds one fat sector at a time  

  ' global master boot record variable template for partition entry (subset, important only)
  long mbr_boot_descriptor            ' boot descriptor (determines if partition entry is active, $80=active, $00=inactive), offset $00, 1-byte
  long mbr_first_part_sect            ' first partition sector, offset $01, 1-byte
  long mbr_file_sys_desc              ' fat file system descriptor ($04 means fat16 system < 32MB, $06 means fat16 > 32MB, offset $4, 1-byte
  long mbr_last_part_sect             ' last partition sector,  offset $05, 3-bytes                  
  long mbr_num_sect_to_part           ' number of sectors between the MBR and the first sector of the partition, offset $08, 4-bytes
  long mbr_num_sect_in_part           ' number of sectors in the partition, offset $0C, 4-bytes

  ' global partition boot sector variable template (subset, important only) 
  long pbr_bytes_per_sect
  long pbr_sect_per_clust
  long pbr_num_res_sect
  long pbr_num_fats
  long pbr_num_root_dir_entrys
  long pbr_total_num_sect
  long pbr_num_sect_per_fat
  long pbr_num_vol_id
  byte pbr_volume_label[12]

  ' global FAT tracking globals      
  long first_data_sector
  long first_dir_sector
  long first_fat_sector

  ' temp global FAT directory entry (subset, important only)
  long dir_attr   
  long dir_reserved 
  long dir_create_time_ms
  long dir_create_hms         
  long dir_create_date_ymd  
  long dir_last_access_date_ymd 
  long dir_ea_index                
  long dir_last_update_time_hms
  long dir_last_update_date_ymd
  long dir_first_cluster       
  long dir_file_size
  long hours, minutes, seconds                  
  long year, month, day

  byte tfilename[16] ' temporary string buffer, used for various string operations by functions internally

  ' END API Globals -----------------------------------------------------------

OBJ

  key  : "GBP_gkey.spin"          ' keyboard object for local debugging
  lcd  : "GBP_MIO283QT2.spin"     ' lcd terminal, note version
  
PUB Start (buffer_ptrs)
' starts the SD driver up, does any global initialization that is needed (not much)
' also this function takes as an input a pointer to a receiver area that the function writes the local addresses of the
' sector buffers, so the calling object can use the memory and not create his own local buffers. the data structure
' is assumed to look like this from the callers point of view:
' long diskbuff_ptr ' generic disk sector buffer
' long mbrbuff_ptr  ' master boot record buffer
' long pbrbuff_ptr  ' partition boot record
'
' This function expects the address of the first pointer @diskbuff_ptr to be sent from caller, then the function wil
' fill in each of the pointers and point them to the local memory buffers, so the caller can access them directly
'
' INPUTS: buffer_ptrs - pointer to starting address of diskbuff_ptr, mbrbuff_ptr, and pbrbuff_ptr in callers VAR space
'
' OUTPUTS: setups buffer_ptrs and returns $00 for success, $FF for failure 
' -----------------------------------------------------------------------------

  ' set global debug level
  debug_level := DEBUG_BRIEF  

  'start keyboard on pins 26,27, when doing local debugging 
  'key.start(26,27)

  'start the LCD-Display
'  lcd.start(LCD_CS, LCD_SCK, LCD_DI, LCD_RES)

  ' assign caller's pointers to local memory buffers, so caller can access them and use them
  long [buffer_ptrs][0] := @diskbuff ' generic disk sector buffer
  long [buffer_ptrs][1] := @mbrbuff  ' master boot record buffer
  long [buffer_ptrs][2] := @pbrbuff  ' partition boot record

' end Start

  ' export LCD terminal interface to higher level object, so caller doens't have to instantiate TV terminal object
  ' crude hack, but what can you do!

PUB lcd_bin( value, digits )
  lcd.bin( value, digits )
  
PUB lcd_out( char )
  lcd.out( char )
       
PUB lcd_dec( value )
  lcd.dec( value )
 
PUB lcd_hex( value, digits )
  lcd.hex( value, digits )
  
PUB lcd_setx( new_x )
  lcd.setx(  new_x )
 
PUB lcd_sety( new_y )
  lcd.sety( new_y )
  
PUB lcd_getx
  return( lcd.getx )
 
PUB lcd_gety
  return( lcd.gety )

PUB lcd_pstring( string_ptr )
  lcd.pstring( string_ptr )

PUB lcd_crlf
'Carriage return
  lcd.crlf

PUB lcd_cr
  lcd.cr

PUB lcd_lf
'Line feed
  lcd.lf

PUB lcd_start(LCD_CS, LCD_SCK, LCD_DI, LCD_RES)
  lcd.start(LCD_CS, LCD_SCK, LCD_DI, LCD_RES)

PUB lcd_clear
' Clears the whole scren
  lcd.clear
{
PUB tvt_erase
  lcd.erase
}

PUB FAT_Print_Directory | sector_index, dir_index, data8, index, index2, hours_12, total_file_size, total_num_files, keycode

' DESCRIPTION: prints out the root directory entries to the terminal screen, see FAT_File_Open for explanation of how this works
' INPUTS: none
' OUTPUTS: returns the total number of files found 
' -----------------------------------------------------------------------------
  lcd.pstring(STRING("SD Card FAT16 Directory Listing..."))

  ' reset aggregate file size
  total_file_size := 0
  total_num_files := 0

  repeat sector_index from first_dir_sector to (first_dir_sector + 32 - 1) 
    ' read the sector in then traverse each directory entry, there are 16 entries
    ' per sector, once we find the file, or we find and empty entry then we can quit
    SD_Read_Sector(sector_index, @diskbuff)
    lcd.sety( lcd.gety )'- 1) ' HACK - for some reason everytime a sector is read, the inner loop is entered/exited another newline is printed!!!
    lcd.crlf
    lcd.pstring(STRING("More.....key 1 to 7 / key 8 = exit"))
      keycode := key.waitforkey
      if (keycode == 8)
        quit
    lcd.clear

    ' check each directory entry 0..15 in this sector
    repeat dir_index from 0*32 to 15*32 step 32
      ' is this the last entry?
      if ( diskbuff[dir_index] == $00)
        ' print final stats
        ' output file size and number of files
        lcd.crlf
        lcd.setx(7)
        lcd.dec(total_num_files)
        lcd.pstring(STRING(" File(s)"))
        lcd.setx(9+7+4)
        lcd.dec(total_file_size)
        lcd.pstring(STRING(" Bytes"))
      
        ' return total number of entries
        return(total_num_files )
       
      ' first file name, test first byte for deleted, empty, or first character of filename
      elseif ( (diskbuff[dir_index] == $E5) or (diskbuff[dir_index] == $2E) or ( diskbuff[ dir_index + $0B ]==$0F ) or ( diskbuff[ dir_index + $0B ]==$08 ))
        ' skip to next entry
        next   

      ' this is a valid entry, two special cases, volume label and root subdirectory
       
      ' now other parts of file
      dir_attr                 := ( diskbuff[ dir_index + $0B ] )
      dir_reserved             := ( diskbuff[ dir_index + $0C ] )
      dir_create_time_ms       := ( diskbuff[ dir_index + $0D ] )
      dir_create_hms           := ( diskbuff[ dir_index + $0E ] ) + ( diskbuff[ dir_index + $0E + $01 ] << 8 )     
      dir_create_date_ymd      := ( diskbuff[ dir_index + $10 ] ) + ( diskbuff[ dir_index + $10 + $01 ] << 8 ) 
      dir_last_access_date_ymd := ( diskbuff[ dir_index + $12 ] ) + ( diskbuff[ dir_index + $12 + $01 ] << 8 ) 
      dir_ea_index             := ( diskbuff[ dir_index + $14 ] ) + ( diskbuff[ dir_index + $14 + $01 ] << 8 )    
      dir_last_update_time_hms := ( diskbuff[ dir_index + $16 ] ) + ( diskbuff[ dir_index + $16 + $01 ] << 8 )
      dir_last_update_date_ymd := ( diskbuff[ dir_index + $18 ] ) + ( diskbuff[ dir_index + $18 + $01 ] << 8 )
      dir_first_cluster        := ( diskbuff[ dir_index + $1A ] ) + ( diskbuff[ dir_index + $1A + $01 ] << 8 )
      dir_file_size            := ( diskbuff[ dir_index + $1C ] ) + ( diskbuff[ dir_index + $1C + $01 ] << 8 ) + ( diskbuff[ dir_index + $1C + $02 ] << 16 )+ ( diskbuff[ dir_index + $1C + $03 ] << 24 )       

      ' use these commented out sections to print out and format other details about the file entry that you might want to display

      {       
      term.pstring(STRING("dir_attr:                "))                 
      term.bin(dir_attr, 8)    
      term.out($0D)
      }

      { 
      term.pstring(STRING("dir_reserved:            "))           
      term.dec(dir_attr)
      term.out($0D)
      }

      { 
      term.pstring(STRING("dir_create_time_ms:      "))   
      term.dec(dir_create_time_ms)
      term.out($0D)
      }

      { 
      term.pstring(STRING("dir_create_hms:          "))               
      hours   := (dir_create_hms & %1111_1000_0000_0000) >> 11 
      minutes := (dir_create_hms & %0000_0111_1110_0000) >> 5
      seconds := (dir_create_hms & %0000_0000_0001_1111) << 1 ' resolution 2 seconds, so multiply by 2 
       
      term.dec(hours)
      term.out(":")
      term.dec(minutes)
      term.out(":")
      term.dec(seconds)
      term.out($0D)
      } 

      { 
      term.pstring(STRING("dir_create_date_ymd:     "))      
      year  := 1980 + (dir_create_date_ymd & %1111_1110_0000_0000) >> 9  
      month := (dir_create_date_ymd & %0000_0001_1110_0000) >> 5
      day   := (dir_create_date_ymd & %0000_0000_0001_1111)  
       
      term.dec(month)
      term.out("/")
      term.dec(day)
      term.out("/")
      term.dec(year)
      term.out($0D)
      }

      {      
      term.pstring(STRING("dir_last_access_date_ymd:")) 
      year  := 1980 + (dir_last_access_date_ymd & %1111_1110_0000_0000) >> 9  
      month := (dir_last_access_date_ymd & %0000_0001_1110_0000) >> 5
      day   := (dir_last_access_date_ymd & %0000_0000_0001_1111)  
       
      term.dec(month)
      term.out("/")
      term.dec(day)
      term.out("/")
      term.dec(year)
      term.out($0D)
      }
       
      { 
      term.pstring(STRING("dir_ea_index:            "))              
      term.dec(dir_ea_index)
      term.out($0D)
      }

      'term.pstring(STRING("dir_last_update_date_ymd:")) 
      year  := 1980 + (dir_last_update_date_ymd & %1111_1110_0000_0000) >> 9  
      month := (dir_last_update_date_ymd & %0000_0001_1110_0000) >> 5
      day   := (dir_last_update_date_ymd & %0000_0000_0001_1111)  
       
      lcd.dec(month)
      lcd.out("/")
      lcd.dec(day)
      lcd.out("/")
      if ((year-2000) < 10)
        lcd.out("0")
      lcd.dec(year-2000)
      lcd.setx(10)
      'term.pstring(STRING("  "))      
      'term.out($0D)
       
      'term.pstring(STRING("dir_last_update_time_hms:")) 
      hours   := (dir_last_update_time_hms & %1111_1000_0000_0000) >> 11 
      minutes := (dir_last_update_time_hms & %0000_0111_1110_0000) >> 5
      seconds := (dir_last_update_time_hms & %0000_0000_0001_1111) << 1 ' resolution 2 seconds, so multiply by 2 

      ' convert to am/pm
      hours_12 := hours // 12
      if (hours_12 == 0)
        hours_12 := 12
      
      lcd.dec(hours_12)
      lcd.out(":")

      ' add a leading 0
      if (minutes < 10)
        lcd.out("0")
        
      lcd.dec(minutes)
      lcd.out(" ")
      'term.out(":")
      'term.dec(seconds)
      lcd.setx(16)
      if ((hours / 12) == 0)
        lcd.pstring(STRING("AM"))
      else
        lcd.pstring(STRING("PM"))
      
      lcd.setx(20)
      ' term.pstring(STRING("  "))
      ' term.out($0D)
      
      { 
      term.pstring(STRING("dir_first_cluster:       "))        
      term.dec(dir_first_cluster)
      term.out($0D)
      }

      ' test for directory entry eventhough we can't traverse, show root directories
      if (dir_attr & $10)
        lcd.pstring(STRING("<DIR>"))
        lcd.setx(28)
      else
        ' normal file
        ' update file tracking
        total_file_size += dir_file_size
        total_num_files++
        ' term.pstring(STRING("dir_file_size:           "))                  
        lcd.dec(dir_file_size)
        lcd.setx(28)
        'term.pstring(STRING("  "))      
        'term.out($0D)

      ' print file name
      repeat index from dir_index to dir_index+7
        ' test for non-space character
        if (diskbuff[index] <> $20)
          lcd.out(diskbuff[index])
        else  
          quit
                   
      ' test for extension           
      if (diskbuff[dir_index+8] <> $20)
        ' put dot for extension
        lcd.out(".")
        
        repeat index from dir_index+8 to dir_index+10
          if (diskbuff[index] <> $20)
            lcd.out(diskbuff[index])
          else  
            quit
            
      lcd.crlf

    ' end repeat dir_index
    lcd.crlf
       
' end FAT_Print_Directory

PUB FAT_File_Open(filename_ptr, file_handle_ptr) | char, file_found,  fnlength, findex, findex2, sector_index, dir_index, dir_entry, string_ptr1, string_ptr2

' DESCRIPTION: scans the root directory and if the file pointed to by the NULL terminated filename filename_ptr
' is found then opens the file for reading and stores a data structure referring to the state of the file
' in file_handle_ptr which is points to a generic "file descriptor" data structure consisting of 4 longs
' in the following format:
' 
' long first_cluster ' first cluster of file
' long filesize      ' size in bytes of file
' long curr_read_pos ' current read position
' long dir_entry     ' the directory index of the entry
'
' thus, the caller must create a zero'ed out 4 long storage area and pass the pointer to this function to work
' properly. This function is complex and the explanation is best read about in the user manual, but a brief
' overview is as follows: the FAT16 file system consists of a master boot record MBR (which would have already been read in)
' the MBR has partition entries that refer to each of the partitions on the drive (SD card in our case). Each partition entry
' points to the actual partition data of the logical drive, so first step is to load the MBR then located the first partition entry
' within it. Each entry is 16-bytes and the most important thing in the partition entry is the first sector of the partion along
' with the number of sectors from the MBR and the first sector of the partition (redundant more or less). Now, with the first
' sector of the partition, we load that sector in and it is the partition boot sector or volume boot sector or FAT16 boot sector,
' all mean the same thing. In the boot sector is all kind of information about the physical nature of the drive, however, in the
' case of SD cards, most of the information is irrelavent since there or no heads, tracks, etc. But, there are some entries like
' total number of sectors, bytes per sector, sectors per cluster that are needed for calculations and change depending on the size
' of the SD card. Next, up is the actual FAT data structure this is a linked list of clusters for each file, this is followed by the
' actual root DIRECTORY of files, each directory entry is 32-bytes and has name, length, etc. of each file. So there are a lot of data
' structures that have to be traversed, add to that, we only cache one sector at a time (512 bytes), and file access becomes quite
' a challenge on the propeller chip since not only is the memory limited, but SPIN has zero support for data structures, so everything
' in an array. In any event, to "open" a file, the following actions are taken; the root file directory is traversed, when/if the requested
' file name is found then the 32-byte file entry is accessed and the first cluster and size are returned in the sent
' "file handle" data structure. Interestingly, all we need is the size of the file and the first cluster to read it in since
' once we have this information then the FAT16 linked list is all we need. Of course, it might be prudent to also return the actual
' directory entry number, since once the file is opened other functions might refer to it, hence, the entry index is returned as well.
' The function FAT_Print_Directory works exactly as this function does, except it simply prints all valid entries
'
' INPUTS: filename_ptr - NULL terminated 8.3 filename to attempt to open (only in root directory)
'         file_handle_ptr - pointer to 4 longs where the file handle information will be stored if successful
' OUTPUTS: Returns length of file (which could be 0), or -1 if file not found.
' -----------------------------------------------------------------------------
 

  ' convert null terminated search filename into same format 8.3 with spaces padding right side
  ' this will make comparision less complex
  ' extract root filename
   
  ' prepare test filename with all spaces
  bytefill(@tfilename, " ", 11)  
  tfilename[11] := 0 ' null terminate for compatibility
   
  ' compute length of filename being tested      
  fnlength := strsize( filename_ptr )
   
  ' parse filename into root and extention with padding
   
  repeat findex from 0 to (fnlength-1)
     ' test for extention
    if (byte[filename_ptr][findex] == ".")

      repeat findex2 from (findex+1) to (fnlength-1)
        tfilename[(findex2 - (findex + 1)) + 8] := ToUpper(byte[filename_ptr][findex2])
      quit  
    else
      tfilename[findex] := ToUpper(byte[filename_ptr][findex]) 

  ' set file not found
  file_found := 0
   
  ' step 1: scan the directory for the file, see if it exists
  ' there are 512 directory entries, each entry is 32 bytes, thus, 32 sectors to
  ' hold the entire root directory, scan thru all of them
   
  repeat sector_index from first_dir_sector to (first_dir_sector + 32 - 1) 
    ' read the sector in then traverse each directory entry, there are 16 entries
    ' per sector, once we find the file, or we find and empty entry then we can quit
    SD_Read_Sector(sector_index, @diskbuff)
   
    ' check each directory entry 0..15 in this sector
    repeat dir_index from 0*32 to 15*32 step 32
   
      ' first file name, test first byte for deleted, empty, or first character of filename
      if (diskbuff[dir_index] == $00)
        ' last directory entry, no need to search any more entries or sectors
        ' exit, file not found
        ' set out loop past index to exit
        sector_index := first_dir_sector + 32
        ' break out of inner loop
        quit 
         
      elseif ( (diskbuff[dir_index] == $E5) or (diskbuff[dir_index] == $2E) or ( diskbuff[ dir_index + $0B ]==$0F ) or ( diskbuff[ dir_index + $0B ]==$08 ))
        ' file deleted, folder entry, long filename, volume entry move on to next entry in inner loop
        next 

      ' else non-empty, deleted, last, thus must be a real entry
      ' at this point, this is a file of some sort, figure out which?

      ' compare the file name to the entry and see if we have a match
      ' entry format 8.3: filename(8): FFFFFxxx Extention(3): EEx 
      ' where the names are LEFT justified and padded with spaces $20 (32) to the right.

      ' first compare filename
      if ( Strncomp( @tfilename[0], @diskbuff[dir_index], 11 ) == 1)
        ' filename found, exit out of loop

        ' set found flag
        file_found := 1

        ' compute absolute directory entry 0 .. 511 
        dir_entry  := (sector_index-first_dir_sector)*32 + dir_index/32 
        
        sector_index := first_dir_sector + 32
        ' break out of inner loop
        quit 
    
    ' end repeat dir_index
    
  ' end repeat sector_index
  
  ' did we find the file?
  if (file_found==0)
    ' file not found, return 0 for error, otherwise length of file
    if (debug_level == DEBUG_BRIEF)
      lcd.crlf
      lcd.pstring( STRING("File not found") )
      lcd.crlf
    return (-1)
  else
    if (debug_level == DEBUG_BRIEF)
      lcd.crlf
      lcd.pstring( STRING("File found at directory entry [") )
      lcd.dec(dir_entry)
      lcd.pstring( STRING("]") )
      lcd.crlf
    
      ' print filename
      repeat findex from 0 to 10
        lcd.out( ToUpper( diskbuff[dir_index+findex] ) )
      lcd.crlf

    ' extract entire data structure incase we want it later 
    dir_attr                 := ( diskbuff[ dir_index + $0B ] )
    dir_reserved             := ( diskbuff[ dir_index + $0C ] )
    dir_create_time_ms       := ( diskbuff[ dir_index + $0D ] )
    dir_create_hms           := ( diskbuff[ dir_index + $0E ] ) + ( diskbuff[ dir_index + $0E + $01 ] << 8 )     
    dir_create_date_ymd      := ( diskbuff[ dir_index + $10 ] ) + ( diskbuff[ dir_index + $10 + $01 ] << 8 ) 
    dir_last_access_date_ymd := ( diskbuff[ dir_index + $12 ] ) + ( diskbuff[ dir_index + $12 + $01 ] << 8 ) 
    dir_ea_index             := ( diskbuff[ dir_index + $14 ] ) + ( diskbuff[ dir_index + $14 + $01 ] << 8 )    
    dir_last_update_time_hms := ( diskbuff[ dir_index + $16 ] ) + ( diskbuff[ dir_index + $16 + $01 ] << 8 )
    dir_last_update_date_ymd := ( diskbuff[ dir_index + $18 ] ) + ( diskbuff[ dir_index + $18 + $01 ] << 8 )
    dir_first_cluster        := ( diskbuff[ dir_index + $1A ] ) + ( diskbuff[ dir_index + $1A + $01 ] << 8 )
    dir_file_size            := ( diskbuff[ dir_index + $1C ] ) + ( diskbuff[ dir_index + $1C + $01 ] << 8 ) + ( diskbuff[ dir_index + $1C + $02 ] << 16 )+ ( diskbuff[ dir_index + $1C + $03 ] << 24 )       

    ' build up open file descriptor in following format, this is the minimum we need to keep track of the open file
    ' long first_cluster ' first cluster of file
    ' long filesize      ' size in bytes of file
    ' long curr_read_pos ' current read position  
    long[ file_handle_ptr ][0] := dir_first_cluster  
    long[ file_handle_ptr ][1] := dir_file_size     
    long[ file_handle_ptr ][2] := 0 ' set file pointer current pos to first byte of file
    long[ file_handle_ptr ][3] := dir_entry   ' the directory index of the entry    

    ' return length of file (could be 0, which is valid)
    return ( dir_file_size )

' end FAT_File_Open

PUB FAT_File_Read(file_handle_ptr, buffer_ptr, count) | fat_sector, fat_offset, bytes_read, index, sector_offset, data_sector, num_bytes_copy, next_cluster, prev_cluster, fp, fp_file_size, fp_curr_pos, fp_end_pos, fp_first_cluster, logical_sector, logical_cluster, physical_cluster, cluster_index

' DESCRIPTION: this function reads count bytes from the file referred to by the file handle and stores them into the sent buffer
' if count is greater than the total number of bytes left in the file, or the current file pointer plus count is greater
' than the length of the file, the function will read to the end of the file and return the ACTUAL number of bytes read always
'
' This function is VERY VERY complex and the explanation is best read about in the user manual, but a brief
' overview is as follows: the FAT16 file system consists of a master boot record MBR (which would have already been read in)
' the MBR has partition entries that refer to each of the partitions on the drive (SD card in our case). Each partition entry
' points to the actual partition data of the logical drive, so first step is to load the MBR then located the first partition entry
' within it. Each entry is 16-bytes and the most important thing in the partition entry is the first sector of the partion along
' with the number of sectors from the MBR and the first sector of the partition (redundant more or less). Now, with the first
' sector of the partition, we load that sector in and it is the partition boot sector or volume boot sector or FAT16 boot sector,
' all mean the same thing. In the boot sector is all kind of information about the physical nature of the drive, however, in the
' case of SD cards, most of the information is irrelavent since there or no heads, tracks, etc. But, there are some entries like
' total number of sectors, bytes per sector, sectors per cluster that are needed for calculations and change depending on the size
' of the SD card. Next, up is the actual FAT data structure this is a linked list of clusters for each file, this is followed by the
' actual root DIRECTORY of files, each directory entry is 32-bytes and has name, length, etc. of each file. So there are a lot of data
' structures that have to be traversed, add to that, we only cache one sector at a time (512 bytes), and file access becomes quite
' a challenge on the propeller chip since not only is the memory limited, but SPIN has zero support for data structures, so everything
' in an array. In any event, after the file has been "opened" with a call to FAT_File_Open the file handle descriptor is used by this
' function as a locator of the actual file data cluster and its length. Since we know the first cluster and the length of the file we
' simple need to read in the data sector by sector of the file. Now, there are a myriad of problems with this. First, each sector is 512
' bytes, so file reads that are greater than 512 bytes will span a sector, also, file reads that straddle sectors will need to read multiple
' sectors. There are typically 32-64 sectors per cluster, which means a single cluster holds 32-64K bytes, thus if a file is larger than
' 32-64K bytes then more than one cluster will have to be accessed. This means accessing the FAT and traversing the linked list of clusters
' for the file as it passes cluster boundaries. This problem is further compounded by the fact that each FAT sector only holds 256 cluster links,
' thus if a file is greater than 256*cluster size this means that multiple FAT sectors have to be loaded!! All this ends up being a
' brutally painful experience to code in SPIN due to its lack of data structure support and of course the propeller has no memory,
' so everything is cached a sector at a time. Anyway, the trick to writing FAT16 (or 32) functions is to think in terms of logical and
' physical values, so you can write a routine that deals with bytes, then one that deals with sectors, then as the fragmented data structures
' on the SD (or disk) are needed they are buffered transparently for you, not fun, no matter how many times your code it!
'
' INPUTS: 
'         file_handle_ptr - pointer to 4 longs of file handle previously opened with call to FAT_File_Open in following format:
 
'             long first_cluster ' first cluster of file
'             long filesize      ' size in bytes of file
'             long curr_read_pos ' current read position
'             long dir_entry     ' the directory index of the entry
'
'         buffer_ptr - pointer to memory where the data will be read, make sure its big enough! 
'         count      - number of bytes to read.
'          
' OUTPUTS: Returns actual number of bytes read
' -----------------------------------------------------------------------------


  ' make sure this is a valid file
  if (file_handle_ptr == 0)
    if (debug_level == DEBUG_BRIEF)
      lcd.crlf
      lcd.pstring( STRING("FAT16 Error: NULL file handle.") )
    return (0) ' null pointer

  ' extract cluster
  fp_first_cluster := long[ file_handle_ptr ][0]

  ' compute fat sector offset that cluster index exists in, each fat sector holds 256 entries (assuming a 512 byte sector)
  ' for example if the cluster index is from 0..255 then we simply load the first FAT sector, 1..511, then the 2nd FAT sector etc.
  fat_offset := fp_first_cluster / 256

  ' compute current fat sector cached
  fat_sector := first_fat_sector + fat_offset    

  if (debug_level == DEBUG_VERBOSE)
    lcd.pstring( STRING("fat offset = ") )
    lcd.dec(fat_offset)
    lcd.crlf

  ' test if first cluster valid
  if ( fp_first_cluster < 2)
    if (debug_level == DEBUG_BRIEF)
      lcd.crlf
      lcd.pstring( STRING("FAT16 Error: Cluster not valid.") )
    return (0) ' file handle error  

 
  fp_file_size  := long[ file_handle_ptr ][1]     
  fp_curr_pos   := long[ file_handle_ptr ][2]
  fp_end_pos    := _Min( fp_curr_pos + count - 1, fp_file_size - 1) 

  ' recompute count now that its been bounded
  count      := fp_end_pos - fp_curr_pos + 1
  bytes_read := count   

  ' test if file pointer is at end of file (-1 or => file_size), if so simply return, user needs to seek to the start
  if ( (fp_curr_pos == -1) or (fp_curr_pos => fp_file_size))
    ' 0 bytes read
    if (debug_level == DEBUG_BRIEF)    
      lcd.crlf
      lcd.pstring( STRING("End Of File.") )
    return(0)  

  ' compute working variables to make reading file easier (if there is such a thing with FAT!)

  ' load the FAT sector that has the cluster link in it, we might have to load multiple FAT sectors to traverse entire FAT
  if (debug_level == DEBUG_VERBOSE)  
    SD_Print_Sector(fat_sector, @diskbuff, 0,64, 16, 0)
  SD_Read_Sector(fat_sector, @fatbuff)

  ' compute logical cluster, simply divide filepointer by size of sector, 0 based value
  logical_sector := fp_curr_pos / pbr_bytes_per_sect
  
  ' now the logical_cluster, this means the 0 based cluster that refers to the cluster holding the first byte refered to
  ' by the current file pointer, fp_curr_pos
  logical_cluster := logical_sector / pbr_sect_per_clust   

  if (debug_level == DEBUG_VERBOSE)
    lcd.crlf
    lcd.pstring( STRING("pbr_bytes_per_sect=") )
    lcd.dec(pbr_bytes_per_sect)
     
    lcd.crlf
    lcd.pstring( STRING("pbr_sect_per_clust=") )
    lcd.dec(pbr_sect_per_clust)
     
    lcd.crlf
    lcd.pstring( STRING("Logical sector=") )
    lcd.dec(logical_sector)
     
    lcd.crlf
    lcd.pstring( STRING("Logical cluster=") )
    lcd.dec(logical_cluster)
    lcd.crlf
    ' key.newkey

  ' traverse FAT cluster array to find the physical cluster we are interested, a file is broken into clusters, which are each n contiguous sectors
  ' but the cluster need NOT be contiguous, thus we must traverse the FAT to find the actual physical cluster the data is in

  ' this is the first cluster, it points to the next cluster, etc.
  prev_cluster := fp_first_cluster
  next_cluster := fp_first_cluster

  if (debug_level == DEBUG_VERBOSE)
    lcd.pstring( STRING("Finding 1st cluster of file...") )
    lcd.crlf
    ' key.newkey

  ' traverse list
  cluster_index := 0
  
  repeat while (cluster_index < logical_cluster )
    if (debug_level == DEBUG_VERBOSE)
      lcd.pstring( STRING(" Clust ") )
      lcd.dec(prev_cluster)

    ' at this point we need to make sure we haven't crossed a FAT sector boundary since we are only caching one sector
    ' of the FAT which is 512 bytes, 2 bytes per entry, thus we need to test the desired cluster entry to determine if its
    ' still in the current FAT sector, if not, load the needed FAT sector, which COULD be non-linear and somewhere else
    ' if user has deleted / created a lot of files on the SD card 
    if ( (( prev_cluster / 256) + first_fat_sector) <> fat_sector )
      ' re-compute proper FAT sector and load to access next 256 FAT cluster entries 
      fat_offset := prev_cluster / 256

      ' compute current fat sector cached
      fat_sector := first_fat_sector + fat_offset    

      ' and finally load FAT sector into cache
      if (debug_level == DEBUG_VERBOSE)      
        SD_Print_Sector(fat_sector, @diskbuff, 0,64, 16, 0)
        
      SD_Read_Sector(fat_sector, @fatbuff)

      if (debug_level == DEBUG_VERBOSE)
        lcd.pstring( STRING("Loading new FAT sector ") )
        lcd.dec(fat_offset)
        lcd.pstring( STRING(" at final sector ") )
        lcd.dec(fat_sector)
        lcd.crlf

    ' always index into the FAT buffer mod 256, we can do this since we know that we have the correct FAT sector already
    next_cluster := word[ @fatbuff ][ prev_cluster // 256 ]
    
    if (debug_level == DEBUG_VERBOSE)
      lcd.pstring( STRING(" -> ") )
      lcd.dec(next_cluster)

    ' test if this is last cluster
    if (next_cluster == $FFFF)
      next_cluster := prev_cluster
      quit

    ' advance link to next cluster   
    prev_cluster := next_cluster

    ' increment logical cluster index
    cluster_index++

  ' finally we have the physical cluster
  physical_cluster := next_cluster

  if (debug_level == DEBUG_VERBOSE)
    lcd.cr
    lcd.pstring( STRING("Logical cluster ") )
    lcd.dec(logical_cluster)
    lcd.pstring( STRING(" at physical ") )
    lcd.dec(physical_cluster)
    lcd.crlf

   key.waitforkey


  ' ready to copy file to buffer, the idea here is to use the logical byte index to find the cluster, sector, and byte on the SD then copy it
  ' this will be done in sector size chunks for speed
  fp := fp_curr_pos

  repeat while (count > 0)  

    ' compute the actual sector holding the data, cluster and file pointer need to be considered
    data_sector := ((fp / SECTOR_SIZE) // pbr_sect_per_clust ) + (first_data_sector + ((physical_cluster-2) * pbr_sect_per_clust) ) 

    if (debug_level == DEBUG_VERBOSE)
      lcd.crlf
      lcd.pstring( STRING("data sector=") )
      lcd.dec(data_sector)
      lcd.crlf
      'key.newkey
       
    ' now read the sector in that the current file pointer is within    
    SD_Read_Sector(data_sector, @diskbuff)

    ' compute offset into current sector (this happens on first sector, last sector is a similar special case) 
    sector_offset := fp // SECTOR_SIZE

    if (debug_level == DEBUG_VERBOSE)
      lcd.pstring( STRING("sector offset=") )
      lcd.dec(sector_offset)
      lcd.crlf
      'key.newkey
       
    
    ' at most we can copy (SECTOR_SIZE - sector_offset) bytes since we can't cross a sector boundary without another sector read 

    ' now copy the data from the sector to the output buffer, sector_size bytes at a time (max)
    ' test for special cases first

    if ( (sector_offset + count) > SECTOR_SIZE)
      ' data crosses a sector boundary, copy the remainder of the sector
      num_bytes_copy := (SECTOR_SIZE - sector_offset)

      if (debug_level == DEBUG_VERBOSE)
        lcd.pstring( STRING(" > sector size ") )
        lcd.crlf
        'key.newkey
       
    else
      ' data doesn't cross another sector boundary, this is it
      num_bytes_copy := count

      if (debug_level == DEBUG_VERBOSE)
        lcd.pstring( STRING(" < sector size ") )
        lcd.crlf
        'key.newkey

    if (debug_level == DEBUG_VERBOSE)
      lcd.pstring( STRING("num_bytes_copy=") )
      lcd.dec(num_bytes_copy)
      lcd.crlf
      'key.newkey

    ' copy the bytes now
    bytemove( buffer_ptr, @diskbuff + sector_offset, num_bytes_copy )

    ' print to screen
    if (debug_level == DEBUG_BRIEF)    
      repeat index from 0 to num_bytes_copy - 1
        lcd.out( byte [buffer_ptr][index])
        
    ' update count and src/dest pointers
    count      -= num_bytes_copy
    buffer_ptr += num_bytes_copy

    ' advance file pointer to next sector always
    fp         += num_bytes_copy

    ' test if we need to advance to next cluster, there are pbr_sect_per_clust sectors per cluster, if the fp is at a cluster
    ' boundary, then this expression will equal 0
    if ( (fp // (pbr_bytes_per_sect * pbr_sect_per_clust))==0)
      if (debug_level == DEBUG_VERBOSE)
        lcd.pstring( STRING("Cluster boundary at fp = ") )
        lcd.dec(fp)
        lcd.crlf
         
      ' advance to next cluster in FAT chain
      physical_cluster := word[ @fatbuff ][ physical_cluster // 256 ]      

      ' NEW CODE ---------------
      ' now test if new physical cluster causes a FAT sector fault and we have to load in a new FAT sector!
      if ( (( physical_cluster / 256) + first_fat_sector) <> fat_sector )
        ' re-compute proper FAT sector and load to access next 256 FAT cluster entries 
        fat_offset := physical_cluster / 256
       
        ' compute current fat sector cached
        fat_sector := first_fat_sector + fat_offset    
       
        ' and finally load FAT sector into cache
        if (debug_level == DEBUG_VERBOSE)        
          SD_Print_Sector(fat_sector, @diskbuff, 0,64, 16,0)
        
        SD_Read_Sector(fat_sector, @fatbuff)

        if (debug_level == DEBUG_VERBOSE)
          lcd.pstring( STRING("Load new FAT sect ") )
          lcd.dec(fat_offset)
          lcd.pstring( STRING(" at final sect ") )
          lcd.dec(fat_sector)
          lcd.crlf
          ' key.newkey
           
      ' END NEW CODE ---------------       

      if (debug_level == DEBUG_VERBOSE)
        lcd.pstring( STRING("Advancing to cluster = ") )
        lcd.dec(physical_cluster)
        lcd.crlf
         
    'key.newkey

  ' write new file position
  long[ file_handle_ptr ][2] := fp

  if (debug_level == DEBUG_BRIEF)
     
    lcd.crlf
    lcd.pstring( STRING("fp = ") )
    lcd.dec(long[ file_handle_ptr ][2])
     
    lcd.pstring( STRING("  size = ") )
    lcd.dec(long[ file_handle_ptr ][1])
     
    lcd.crlf
     
  ' return number of bytes read
  return ( bytes_read )
 
' end FAT_File_Read

PUB FAT_File_Close(file_handle_ptr)

' DESCRIPTION: simply "closes" the file by zero'ing out the file handle data structure which once again looks like:
'
' file descriptor in following format
' long first_cluster ' first cluster of file
' long filesize      ' size in bytes of file
' long curr_read_pos ' current read position
' long dir_entry     ' the directory index of the entry
'
' INPUTS: file_handle_ptr - pointer to file handle
'
' OUTPUTS: returns 1 if successful, 0 if failed or file handle NULL 
' -----------------------------------------------------------------------------

  ' test pointer, some modicum of error handling
  if (file_handle_ptr == 0)
    return ( 0 ) ' there was a problem, this was null already
   ' else...   
  long[ file_handle_ptr ][0] := 0  
  long[ file_handle_ptr ][1] := 0     
  long[ file_handle_ptr ][2] := 0
  long[ file_handle_ptr ][3] := 0  

  ' return success
  return ( 1 )

' end FAT_File_Close 

PUB FAT_File_Seek(file_handle_ptr, count, mode)

' DESCRIPTION: this function updates the current read position of the sent file handle either
' with a new absolute value, or an offset to advance the pointer, controlled by mode (absolute or relative)
' file descriptor in following format as usual:
' long first_cluster ' first cluster of file
' long filesize      ' size in bytes of file
' long curr_read_pos ' current read position
' long dir_entry     ' the directory index of the entry
'
' INPUTS: file_handle_ptr - file handle to update
'         count           - value to set file position to
'         mode            - mode of operation relative (FILE_SEEK_REL) or absolute(FILE_SEEK_ABS)
'
' OUTPUTS: returns new file position
' -----------------------------------------------------------------------------

  ' test file handle
  if (file_handle_ptr == 0)
    return (-1)
    
  ' test which mode user is requesting
  if (mode == FILE_SEEK_ABS)
    ' update current file position to the minimum of the requested value and the max size
    ' of file - 1
    long[ file_handle_ptr ][2] := _Min(count, long[ file_handle_ptr ][1] - 1)
       
  elseif (mode == FILE_SEEK_REL)
    ' requesting relative mode, so add count to current position, test for overflow
    ' if so, set file read position to end of file
    long[ file_handle_ptr ][2] += count
    if ( long[ file_handle_ptr ][2] => long[ file_handle_ptr ][1])
      long[ file_handle_ptr ][2] := long[ file_handle_ptr ][1] - 1           
  elseif (mode == FILE_SEEK_END)
    ' set file pointer to last byte of file
    long[ file_handle_ptr ][2] := (long[ file_handle_ptr ][1] - 1)    
  else
    ' assume FILE_SEEK_START, set file pointer to first byte of file
    long[ file_handle_ptr ][2] := 0

  ' return new fileposition
  return( long[ file_handle_ptr ][2] )

' end FAT_File_Seek(file_handle_ptr, count, mode) 

PUB FAT_Print_Partition_Boot_Rec | index, data8

' DESCRIPTION: this function pretty prints the loaded partition boot record out in human readable form
' it assumes the partition information has been loaded already with a call to
' FAT_Load_Partition_Boot_Rec
'
' INPUTS: none
' OUTPUTS: prints to terminal
' ----------------------------------------------------------------------------- 

  lcd.pstring(STRING("Partition Boot Record Entry:"))
  lcd.crlf

  lcd.pstring(STRING("pbr_bytes_per_sect = "))
  lcd.dec(pbr_bytes_per_sect)
  lcd.crlf

  lcd.pstring(STRING("pbr_sect_per_clust = "))
  lcd.dec(pbr_sect_per_clust)
  lcd.crlf

  lcd.pstring(STRING("pbr_num_res_sect = "))
  lcd.dec(pbr_num_res_sect)
  lcd.crlf

  lcd.pstring(STRING("pbr_num_fats = "))
  lcd.dec(pbr_num_fats)
  lcd.crlf

  lcd.pstring(STRING("pbr_num_root_dir_entrys = "))
  lcd.dec(pbr_num_root_dir_entrys)
  lcd.crlf

  lcd.pstring(STRING("pbr_total_num_sect = "))
  lcd.dec(pbr_total_num_sect)
  lcd.crlf

  lcd.pstring(STRING("pbr_num_sect_per_fat = "))
  lcd.dec(pbr_num_sect_per_fat)
  lcd.crlf

  lcd.pstring(STRING("pbr_volume_label = "))
  lcd.pstring(@pbr_volume_label)
  lcd.crlf

' end FAT_Print_Partition_Boot_Rec 

PUB FAT_Load_Partition_Boot_Rec(pbr_ptr) | index

' DESCRIPTION: this function reads in the partition boot record (aka volume record, fat boot record),
' this is different than the MBR master boot record in that THIS is the boot record/entry for a
' specific partition, in this case, the partition previously loaded in via the load FAT_Load_Partition_Entry
' call which simply loads one of the 4 possible partition entries in the MBR, each entry is 16 bytes, these
' partition entries help locate the actual partition boot record which has all the information for the physical
' partition. This function loads the data and then parses it apart and sets the global template variables
' 
' INPUTS: pbr_ptr - pointer to the buffer to receive the partition boot record data
' OUTPUTS: assigns the global template variables pbr_* with the values found in the partition record
' -----------------------------------------------------------------------------

  ' read partition boot record, its located mbr_num_sect_to_part sectors from sector 0 (the MBR) 
  SD_Read_Sector(mbr_num_sect_to_part, pbr_ptr)

  if (debug_level == DEBUG_BRIEF)
    lcd.pstring(STRING("Loading partition boot record from "))
    lcd.dec(mbr_num_sect_to_part)
    lcd.crlf
  ' now process record and parse out variables and store to globals
  ' extract partition boot record values into template for easy access
  pbr_bytes_per_sect := ( byte[pbr_ptr][ $0B ] ) + ( byte[pbr_ptr][ $0B + $01 ] << 8 )
  pbr_sect_per_clust := ( byte[pbr_ptr][ $0D ] )
  pbr_num_res_sect   := ( byte[pbr_ptr][ $0E ] ) + ( byte[pbr_ptr][ $0E + $01 ] << 8 ) 
  pbr_num_fats       := ( byte[pbr_ptr][ $10 ] ) 
  pbr_num_root_dir_entrys := ( byte[pbr_ptr][ $11 ] ) + ( byte[pbr_ptr][ $11 + $01 ] << 8 )

  ' check for 32MB partition?
  if ( ( byte[pbr_ptr][ $13 ]==0 ) and ( byte[pbr_ptr][ $13 + $01 ] )==0 )
    ' parition greater than 32M
    pbr_total_num_sect    := ( byte[pbr_ptr][ $20 ] ) + ( byte[pbr_ptr][ $20 + $01 ] << 8 ) + ( byte[pbr_ptr][ $20 + $02 ] << 16) + ( byte[pbr_ptr][ $20 + $03 ] << 24 )  
  else
    ' parition less than 32M
    pbr_total_num_sect    := ( byte[pbr_ptr][ $13 ] ) + ( byte[pbr_ptr][ $13 + $01 ] << 8 ) 

  pbr_num_sect_per_fat    := ( byte[pbr_ptr][ $16 ] ) + ( byte[pbr_ptr][ $16 + $01 ] << 8 )
  pbr_num_vol_id          := ( byte[pbr_ptr][ $27 ] ) + ( byte[pbr_ptr][ $27 + $01 ] << 8 ) + ( byte[pbr_ptr][ $27 + $02] << 16) + ( byte[pbr_ptr][ $27 + $03 ] << 24 ) 

  bytemove( @pbr_volume_label, @pbrbuff[ $2B ], 11)
  pbr_volume_label[11] := 0 ' null terminate

  ' convert any non-printable characters to ascii "."
  repeat index from 0 to 10
    if (pbr_volume_label[index] < $20 or pbr_volume_label[index] > $7E)
      pbr_volume_label[index] := "."        

  ' compute location of relevant data FAT directory structures
  first_fat_sector  := mbr_num_sect_to_part + pbr_num_res_sect
  first_dir_sector  := mbr_num_sect_to_part + pbr_num_res_sect + pbr_num_fats*pbr_num_sect_per_fat   
  first_data_sector := mbr_num_sect_to_part + pbr_num_res_sect + pbr_num_fats*pbr_num_sect_per_fat + (pbr_num_root_dir_entrys / 16) 

  if (debug_level == DEBUG_BRIEF)
     
    lcd.crlf
    lcd.pstring(STRING("first_fat_sector  = "))
    lcd.dec(first_fat_sector)
    lcd.crlf
     
    lcd.pstring(STRING("first_dir_sector  = "))
    lcd.dec(first_dir_sector)
    lcd.crlf
     
    lcd.pstring(STRING("first_data_sector = "))
    lcd.dec(first_data_sector)
    lcd.crlf
     
' end FAT_Load_Partition_Boot_Rec

PUB FAT_Load_Partition_Entry(partition, mbr_ptr) | part_base

' DESCRIPTION: parses the sent mbr buffer and requested partition entry (0..3) and loads global variables
' with the partition entry, further FAT calls will use this data to locate files
' you can change to any partition at any time, by loading in another partition entry
' of course partition 1 us usually all that SD support
'
' INPUTS: partition - integer index of entry to load 0..3 (99% of the time will be 0)
'         mbr_ptr   - pointer to previously loaded master boot record
'
' OUTPUTS: parses and assigns global template variables mbr_*
'          returns success 1, or failure 0
' -----------------------------------------------------------------------------

  ' test mbr_ptr
  if (mbr_ptr == 0)
    return (0) ' failure

  ' compute base address of 16 byte partition record
  part_base := $1BE + partition*16

  ' compute all globals associated with MBR
  
  ' boot descriptor (determines if partition entry is active, $80=active, $00=inactive), offset $00, 1-byte
  mbr_boot_descriptor := ( byte[mbr_ptr][ $04 + part_base ] )            

  ' first partition sector, offset $01, 1-byte
  ' 3-bytes in little endian format, build up LONG from bytes
  mbr_first_part_sect := ( byte[mbr_ptr][ $01 + part_base ] ) + ( byte[mbr_ptr][ $01 + part_base ] << 8 ) + ( byte[mbr_ptr][ $01 + part_base ] << 16 )          

  ' fat file system descriptor ($04 means fat16 system < 32MB, $06 means fat16 > 32MB, offset $4, 1-byte
  mbr_file_sys_desc := ( byte[mbr_ptr][ $04 + part_base ] )              

  ' last partition sector,  offset $05, 3-bytes
  ' 3-bytes in little endian format, build up LONG from bytes  
  mbr_last_part_sect := ( byte[mbr_ptr][ $05 + part_base ] ) + ( byte[mbr_ptr][ $06 + part_base ] << 8 ) + ( byte[mbr_ptr][ $07 + part_base ] << 16 )                               

  ' number of sectors between the MBR and the first sector of the partition, offset $08, 4-bytes
  ' extract number of sectors to partition 1 data from MBR, 4-bytes in little endian format, build up LONG from bytes
  mbr_num_sect_to_part := ( byte[mbr_ptr][ $08 + part_base ] ) + ( byte[mbr_ptr][ $09 + part_base ] << 8 ) + ( byte[mbr_ptr][ $0A + part_base ] << 16 ) + ( byte[mbr_ptr][ $0B + part_base ] << 24 )  

  ' return success
  return (1)

' end FAT_Load_Partition_Entry


PUB FAT_Read_MBR(mbr_ptr)

' DESCRIPTION: simply reads in sector 0 which is the MBR or master boot record, this could be done with the SD_Read_Sector               
' function as well.
'
' INPUTS: mbr_ptr - sector pointer to where the data should be stored (one sector 512 bytes)
' OUTPUTS: returns $FF if there was an error or $00 if successful

  ' read the sector in and return result of read
  return ( SD_Read_Sector(0, mbr_ptr) )

' end FAT_Read_MBR 

PUB FAT_Print_Partition_Entry(partition, mbr_ptr) | part_base, index, data8, _mbr_boot_descriptor, _mbr_first_part_sect, _mbr_file_sys_desc, _mbr_last_part_sect, _mbr_num_sect_to_part

' DESCRIPTION: This funtion simply prints the 16 bytes that make up a partition entry in a tabular hex format "index:[value]"
' this function does NOT load the partition entry into the globals, it just prints them out for inspection
' typically SD cards only use the first first partition entry, but the function support printing all 4 partition entries 0..3
' for completeness. This function is mostly for debugging purposes and not needed, it just helps when you are trying to understand
' what's in the partition entries and to print them to the screen in hex. You must have previously loaded the MBR (sector 0) into
' mbr_ptr before calling this function.

' INPUTS: partition  - 0..3 the partition to print
'         mbr_ptr - pointer to local buffer holding mbr
'
' OUTPUTS: Just prints to screen for inspection.
' -----------------------------------------------------------------------------
 
  ' compute base address of 16 byte partition record
  part_base := $1BE + partition*16

  lcd.pstring(STRING("Partion entry "))
  lcd.dec(partition)
  lcd.pstring(STRING(" at address $"))
  lcd.hex(part_base, 4)
  lcd.crlf
    
  ' print out the requested partition entry, 16 bytes, starts at $1BE
  repeat index from 0 to 15
    lcd.hex(index,2)
    lcd.pstring(STRING(":["))
    data8 := byte[mbr_ptr][ index + part_base ] 
    lcd.hex (data8, 2)
    lcd.pstring(STRING("] "))
    if ((index // 4)==3)
      lcd.crlf

  ' now print out parsed data from partition entry in human readable form

    ' compute base address of 16 byte partition record
  part_base := $1BE + partition*16

  ' compute some locals for printing, use same name as globals, but with underscores
  
  ' boot descriptor (determines if partition entry is active, $80=active, $00=inactive), offset $00, 1-byte
  _mbr_boot_descriptor := ( byte[mbr_ptr][ $04 + part_base ] )            

  ' first partition sector, offset $01, 1-byte
  ' 3-bytes in little endian format, build up LONG from bytes
  _mbr_first_part_sect := ( byte[mbr_ptr][ $01 + part_base ] ) + ( byte[mbr_ptr][ $01 + part_base ] << 8 ) + ( byte[mbr_ptr][ $01 + part_base ] << 16 )          

  ' fat file system descriptor ($04 means fat16 system < 32MB, $06 means fat16 > 32MB, offset $4, 1-byte
  _mbr_file_sys_desc := ( byte[mbr_ptr][ $04 + part_base ] )              

  ' last partition sector,  offset $05, 3-bytes
  ' 3-bytes in little endian format, build up LONG from bytes  
  _mbr_last_part_sect := ( byte[mbr_ptr][ $05 + part_base ] ) + ( byte[mbr_ptr][ $06 + part_base ] << 8 ) + ( byte[mbr_ptr][ $07 + part_base ] << 16 )                               

  ' number of sectors between the MBR and the first sector of the partition, offset $08, 4-bytes
  ' extract number of sectors to partition 1 data from MBR, 4-bytes in little endian format, build up LONG from bytes
  _mbr_num_sect_to_part := ( byte[mbr_ptr][ $08 + part_base ] ) + ( byte[mbr_ptr][ $09 + part_base ] << 8 ) + ( byte[mbr_ptr][ $0A + part_base ] << 16 ) + ( byte[mbr_ptr][ $0B + part_base ] << 24 )  

  lcd.pstring(STRING("mbr_boot_descriptor = "))
  lcd.hex(_mbr_boot_descriptor, 2)
  lcd.crlf

  lcd.pstring(STRING("mbr_first_part_sect = "))
  lcd.dec(_mbr_first_part_sect)
  lcd.crlf

  lcd.pstring(STRING("mbr_file_sys_desc = $"))
  lcd.dec(_mbr_file_sys_desc)
  lcd.crlf

  lcd.pstring(STRING("mbr_last_part_sect = "))
  lcd.dec(_mbr_last_part_sect)
  lcd.crlf

  lcd.pstring(STRING("mbr_num_sect_to_part = "))
  lcd.dec(_mbr_num_sect_to_part)
  lcd.crlf

' end FAT_Print_Partition_Entry 


PUB SPI_Select_Channel( channel )
{
This function sets the active SPI channel chip select on the SPI mux

  SPI_SEL 0 1 2 Channel
          A B C Y
          0 0 0 0
          1 0 0 1
          0 1 0 2
          1 1 0 3
          0 0 1 4
          1 0 1 5
          0 1 1 6
          1 1 1 7
PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 2MB FLASH Memory Bank 0.
          4 - 2MB FLASH Memory Bank 1.
          5 - SD Card, SD-Modul.
          6 - SD-Modul.
          7 - SD-Modul.

RETURNS: nothing.

}

  case channel
    0 : outa[0..2] := %000

    1 : outa[0..2] := %100

    2 : outa[0..2] := %010

    3 : outa[0..2] := %110

    4 : outa[0..2] := %001

    5 : outa[0..2] := %101

    6 : outa[0..2] := %011

    7 : outa[0..2] := %111


' end SPI_Select_Channel

PUB SPI_Init(mode)

' DESCRIPTION: this function initializes the hardware interface to the SPI bus. on the HYDRA SD MAX card
' (as with most SPI buses) there are only 4 signals; clk, data in, data out, and chip select

' INPUTS: mode - c ontrols the phase of the clock and the logic levels
' mode 0 - default support 
' mode 1 - not supported
' mode 2 - not supported
' mode 3 - not supported
'
' OUTPUTS: none 
' -----------------------------------------------------------------------------
{{
This function initializes the SPI IO's.

PARMS: none.
RETURNS: nothing.
}}
  ' set direction to output and low ==> Channel 0
  DIRA[0..2]~~               ' set to output
  OUTA[0..2]         := %000 ' set low

  ' set up SPI lines
  DIRA[SPI_MOSI]     := 1    ' set to output
  DIRA[SPI_MISO]     := 0    ' set to input
  DIRA[SPI_SCK]      := 1    ' set to output

  OUTA[SPI_MOSI]     := 0    ' set to LOW
  OUTA[SPI_SCK]      := 0    ' set to LOW


  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init

PUB SPI_Send_Recv_Byte(spi_data8) | spi_result8, index

' DESCRIPTION: this functions sends spi_data8 and receives spi_result8, remember all spi operations are circular, so
' when a byte is sent a byte is received at the same time, you can ignore the sent or received byte as you
' wish. For example, to do a read just send the dummy value $FF
' NOTE: assumes caller is controlling the chip select line
'
' INPUTS: spi_data8 - 8 bit data that is going to be sent
' OUTPUTS: returns the 8-bit data received at the same time
' -----------------------------------------------------------------------------

  ' reset result
  spi_result8 := 0

  ' shift 8-bits of data out while shifting in 8-bits of data
  ' data is sent and receiving MSB to LSB, so bit 7,6,5,4,3,2,1,0
  ' assume SPI interface is in mode 0/1, later make more flexible 
  repeat index from 0 to 7
    ' place next bit on data in pin to device
    OUTA [ spiDI ] := ((spi_data8 & $80) >> 7)
'   lcd.dec((spi_data8 & $80) >> 7)
    
    spi_data8 <<= 1 ' shift data to left and prepare next bit for transmission

  ' pulse the clock line
    OUTA [ spiCLK ] := 1

      ' data is ready now on output line from device
    spi_result8 <<= 1            ' shift result to left
    spi_result8 |= INA [ spiDO ] ' read data bit and OR into result
                                    
    ' finish clock pulse
    OUTA [ spiCLK ] := 0

'  lcd.pstring(STRING("RECEIVED = "))
'  lcd.hex(spi_result8, 2)
'  lcd.cr

  ' return result
  return (spi_result8)

' end SPI_Send_Receive_Byte

PUB SPI_Read_Byte

' DESCRIPTION: reads a single byte from the SPI interface, simply a dummy call to SPI_Send_Recv_Byte
'
' INPUTS: none. 
'
' OUTPUTS: returns the byte read from the SPI interface
' -----------------------------------------------------------------------------

  return ( SPI_Send_Recv_Byte($FF) )

' end SPI_Read_Byte

PUB SPI_Write_Byte(spi_data8)

' DESCRIPTION: sends a single byte to the SPI interface, returns the read value
'
' INPUTS: spi_data8 - 8-bit data to send 
'
' OUTPUTS: returns the 8-bit data from the SPI interface
' -----------------------------------------------------------------------------

' return the data from the cylic buffer regardless...
return( SPI_Send_Recv_Byte(spi_data8) )

' end SPI_Write_Byte

PUB SD_Read_WP
' DESCRIPTION: reads the single bit "write protect" signal on the SD card mechanical
'
' INPUTS: none 
'
' OUTPUTS: returns write protect bit, 0=write enabled, 1=write protected
' -----------------------------------------------------------------------------
 
  DIRA [ WP ]  := 0 ' set to LOW for input (should be default)
  return ( INA [ WP ] )

' end PUB SD_ReadWP


PUB SD_Read_CD
' DESCRIPTION: reads the single bit "card inserted" signal on the SD card mechanical
'
' INPUTS: none 
'
' OUTPUTS: returns card inserted bit, 0=card inserted, 1=card not inserted
' -----------------------------------------------------------------------------

  DIRA [ CD ]  := 0 ' set to LOW for input (should be default)
  return ( INA [ CD ] )

' end PUB SD_Read_CD

PUB SD_Send_Command( command8, address32 ) | response, index

' DESCRIPTION: This function is the main workhorse of the SD interface and is used to send commands
' to the SD card via the SPI interface. All SD commands consist of 6 bytes in the following format:
'
' Command id     (1-byte) This you get from the SD specification.
' 32-bit address (4-bytes) In big endian form.
' CRC result     (1-byte) Error checking byte.
' 
' The address is in big endian form, that is high byte to low byte (opposite of PC and propeller)
' The cycle redundancy check byte is sent in CRC7 or CRC16 format for all transactions
' except status tokens. CRC is of course a method of error detection, but not
' correction. If a CRC fails to match then the data must be sent or requested again.
' the algorithms for CRC7 and CRC16 are well documented and are basically division algorithms
' where the CRCs are the remainders. In our case, we are always going to send $95 which
' is the CRC for the first command that will be sent that matters, after that CRCs are
' ignored in SPI mode, later add code to compute the CRC for kicks if you like.
' Also, note the use of SD_RESPONSE_LIMIT, this allows the command to be sent over and over until there is a response.
' but, if the attempts exceed this constant, there is something wrong with the SD card.
'
' INPUTS: command8 - 8-bit command from SD card specification refer to document "SD Card Specification" from SD Association.
'         address32 - 32-bit address in normal little endian format, function converts to big endian before transport
'
' OUTPUTS: returns status of command to SD card, typically $FF means error, but for each command response can be different
' -----------------------------------------------------------------------------

  ' first send command
  SPI_Send_Recv_Byte(command8)  

  ' now send 4-bytes of address in MS-Byte to LS-byte format
  SPI_Send_Recv_Byte( (address32 >> 24) & $FF)  
  SPI_Send_Recv_Byte( (address32 >> 16) & $FF)
  SPI_Send_Recv_Byte( (address32 >> 8)  & $FF)
  SPI_Send_Recv_Byte( (address32 >> 0)  & $FF) 

  ' finally the CRC
  SPI_Send_Recv_Byte( $95 )            

  ' now wait for response
  response := $FF ' default response is $FF which means error or time out

  repeat index from 0 to SD_RESPONSE_LIMIT
    ' read response by sending dummy data
    response := SPI_Send_Recv_Byte( $FF )
  
    ' DEBUG CODE - send out number of iterations the response took to get back
    debug_wait_res_num_iter := index

    ' test if we have a non-$FF response
    if (response <> $FF)
      ' send dummy data to allow SD card to finish operation, needed after all commands
      SPI_Send_Recv_Byte ( $FF )
      ' return results
      return ( response )
      
  ' if we got here then something bad happened!
  return ( $FF )

' end SD_Send_Command(command8, address32)

PUB SD_Print_Sector(sector32, sect_ptr, start_byte, end_byte, base, print_addr) | index, data8

' DESCRIPTION: prints out the requested sector from start_byte to end_byte inclusive in various base formats
' for debugging purposes. In decimal and hex modes the printing is done in a tabular form with the starting address
' of the row printed to the left, this address usually would be 0 based, but if the caller wants to offset
' it by a constant value, then print_addr is used for this purpose, print_addr is added to the
' current printed address that is displayed. Also if sector32 == -1, then the function prints the
' data already pointed to by sect_ptr
'
' INPUTS: sector32 - sector to print out, if -1 then function is used to pretty print the sent buffer
'         sect_ptr - buffer space that will be used by function to load sector to print it, will be overwritten 
'         start_byte - 0 based start byte index to begin printing
'         end_byte   - last byte index to print
'         base       - numerical base to print in, 16 for hex, 10 for decimal format, 127 for ASCII
'         print_addr - used to offset the left column base address in decimal and hex tabular forms, set to 0 usually 
'
' OUTPUTS: prints to terminal screen
' -----------------------------------------------------------------------------

  ' read sector into buffer
  if (sector32 <> -1)
    SD_Read_Sector(sector32, sect_ptr)  
  ' else just print data pointed to by sect_ptr

  ' test for ASCII format, if so just print it

  if (base == 127)
    repeat index from 0 to (end_byte - start_byte)
      ' ascii format
      lcd.out( byte[sect_ptr][ start_byte + index ] )
    lcd.crlf
    return
  
  ' print requested data out in tabular decimal or hex  
  repeat index from 0 to (end_byte - start_byte)

    if ((index // 8)==0)
      lcd.crlf

      if (base == 16)
        lcd.pstring(STRING("$"))
        lcd.hex(print_addr + start_byte + index,4)
      else
        lcd.dec(print_addr + start_byte + index)

      lcd.pstring(STRING(": "))

    ' print data now
    data8 := byte[sect_ptr][ start_byte + index ] 

    if (base == 16)
      ' print in hex format
      lcd.pstring(STRING("$"))
      lcd.hex(data8,2)
    else
      ' decimal format
      lcd.dec(data8)

    lcd.pstring(STRING(","))

  lcd.crlf

PUB SD_Read_Sector(sector32, sectorbuffer16) | response, index

' DESCRIPTION: this function reads a sector from the SD card into the sent sector buffer
'
' INPUTS: sector32     - 32-bit absolute sector number to read
'         sectorbuffer - 16-bit pointer to receive buffer for sector 
'
' OUTPUTS: storage pointed to by sectorbuffer16 will be filled with byte data from sector
'          returns $00 if no errors, $FF if error
' -----------------------------------------------------------------------------

  ' send the read block command with the byte address computed by multiplying the
  ' requested sector by 512 (512 bytes per sector)
  response := SD_Send_Command( SD_READ_SINGLE_BLOCK , sector32 * SECTOR_SIZE )

  ' test response, should be $00
  if (response <> $00)
    return (response)

  ' wait for the SD card to fetch the data from the flash storage
  repeat index from 0 to SD_RETRIEVE_SECTOR_LIMIT
    response := SPI_Send_Recv_Byte($FF)

    ' test for proper response code $FE, break out of loop
    if (response == $FE)
      quit
       
    ' // end repeat index loop

  ' return error if not responding    
  if (response <> $FE)
    return ($FF) ' there was an error 

  ' now, read the data bytes into the sent buffer
  repeat index from 0 to 511
    byte[sectorbuffer16][index] := SPI_Send_Recv_Byte($FF)    

  ' next read in dummy data  
  SPI_Send_Recv_Byte($FF) ' ignore this data
  SPI_Send_Recv_Byte($FF) ' ignore CRC value

  SPI_Send_Recv_Byte($FF) ' finally 8 more clocks for the card to finish

  ' return $00, no errors code  
  return ($00)             

' end SD_Read_Sector

PUB SD_Write_Sector(sector32, sectorbuffer16) | response, index

' DESCRIPTION: send the write block command with the byte address computed by multiplying the
' requested sector by 512 (512 bytes per sector). Be CAREFUL, you can destroy the MBR etc. by
' writing and some SD cards can stop functioning, so know what you're doing if you use the write
' operation.
'
' INPUTS: sector32 - sector to write
'         sectorbuffer16 - pointer to sector to write (512 bytes)
' 
' OUTPUTS: returns response code from write operation, $00 no error, $FF error
' -----------------------------------------------------------------------------

  response := SD_Send_Command( SD_WRITE_BLOCK , sector32 * SECTOR_SIZE )

  ' test response, should be $00
  if (response <> $00)
    return (response)

  ' send $FF and $FE
  SPI_Send_Recv_Byte($FF)
  SPI_Send_Recv_Byte($FE)

  ' now, write the bytes to card from buffer
  repeat index from 0 to 511
    SPI_Send_Recv_Byte(byte[sectorbuffer16][index])    

  ' next read in dummy data  
  SPI_Send_Recv_Byte($FF) ' ignore this data
  SPI_Send_Recv_Byte($FF) ' ignore CRC value

  ' wait for write to complete
  response := SD_Wait_Write_Complete

  SPI_Send_Recv_Byte($FF) ' finally 8 more clocks for the card to finish

  ' return response from wait call, $00 = no error, $FF = error  
  return (response)             

' end SD_Write_Sector

PUB SD_Wait_Write_Complete | num_attempts, response

' DESCRIPTION: This function is called after a write operation to "wait" for the write to complete. As
' with any flash technology the write operation can be many times slower than read, thus you must wait for
' it to complete. The function waits SD_WRITE_SECTOR_LIMIT iterations by checking the status response from
' the SD card, if the proper response is given, then its assumed there was a failure. 
'
' INPUTS: none, a previous sector write is assumed 
'
' OUTPUTS: returns $00 for success, $FF for failure
' -----------------------------------------------------------------------------

  num_attempts := 0
  repeat while (num_attempts++ < SD_WRITE_SECTOR_LIMIT )
    ' wait for response
    response := SPI_Send_Recv_Byte($FF)   

    ' looking for $05 for success, $0B or $0D problem
    if ( response == $05)
      quit
    elseif( response == $0B or response == $0D)
      return $FF

  ' test for took too long
  if ( num_attempts => SD_WRITE_SECTOR_LIMIT)
    return $FF

  ' finally wait for first non-$00 response, then done
  num_attempts := 0
  repeat while( (num_attempts++ < SD_WRITE_SECTOR_LIMIT) and (SPI_Send_Recv_Byte($FF) == $00) )

  ' did it take too long to get a response
  if ( num_attempts < SD_WRITE_SECTOR_LIMIT)
    return $00
  else
    return $FF

' end SD_Wait_Write_Complete

PUB SD_Unmount

' DESCRIPTION: unmounts the SD card, simply releases the SD card chip select line.
' In a more advanced file system there might be numerous data structure that are created when
' a file system is "mounted", but in our case there are none, so nothing to unmount!
'
' INPUTS: none 
'
' OUTPUTS: none
' -----------------------------------------------------------------------------

  if (debug_level == DEBUG_BRIEF)
    lcd.pstring(STRING("SD card unmounted"))
    lcd.crlf

  ' de-assert chip select
  'OUTA [ spiCS ]  := 1
  SPI_Select_Channel( 0 )

PUB SD_Mount | index, index2, sd_response, cmd_num_attempts

' DESCRIPTION: This function mounts the SD card which is a fairly complex process consisting of a number of steps.
' 1. Power up.
' 2. SD card must be placed into "idle state"
' 3. Then taken out of idle state.
' 4. Next, the card is tested to make sure its an SD and not MMC.
' 5. Finally, if all passes then the card is ready to go and "mounted".
'
' INPUTS: none
 '
' OUTPUTS: returns $00 for success, $FF for failure 
' -----------------------------------------------------------------------------

  if (debug_level == DEBUG_BRIEF)
    'lcd.sety (lcd.gety + 1)
    lcd.pstring(STRING("Mounting SD card..."))
    lcd.crlf

   'reset SPI- channel to 0
  SPI_Select_Channel( 0 )

  
  ' now clock the SD SPI interface a few times and let it initialize
  ' (send dummy data)
  repeat index from 0 to SD_POWERUP_LIMIT
    SPI_Send_Recv_Byte( $FF )

  ' select SPI-channel 5, SD-Card
  SPI_Select_Channel( 5 )

  ' test for verbose mode
  if (debug_level == DEBUG_VERBOSE)
    lcd.pstring(STRING("Placing SD card into idle state..."))
    lcd.crlf
  
  sd_response := $FF
  cmd_num_attempts := 0
  repeat while (sd_response <> $01)

    ' send SD command GO_IDLE_STATE = $40, which resets the SD card
    sd_response := SD_Send_Command( SD_GO_IDLE_STATE, 0 )  

    ' test for verbose mode
    if (debug_level == DEBUG_VERBOSE)
      ' print response
      lcd.pstring(STRING("Response = "))
      lcd.hex(sd_response, 2)
      lcd.crlf
       
      lcd.pstring(STRING("Number of iterations = "))
      lcd.dec(debug_wait_res_num_iter)
      lcd.crlf
       
      lcd.pstring(STRING("Number of attempts = "))
      lcd.dec(cmd_num_attempts)
      lcd.crlf
       
    ' check if we have tried this long enough
    if (++cmd_num_attempts > SD_RESPONSE_LIMIT)
      ' test for verbose mode
      if (debug_level == DEBUG_VERBOSE)
        lcd.pstring(STRING("SD card not responding to idle state command."))
        lcd.crlf

      if (debug_level == DEBUG_BRIEF)
        lcd.pstring(STRING("Couldn't mount SD card!"))
        lcd.crlf

      return ($FF)

  ' test for verbose mode
  if (debug_level == DEBUG_VERBOSE)
    lcd.pstring(STRING("Bringing SD card out of idle state..."))
    lcd.crlf
     
  sd_response := $FF
  cmd_num_attempts := 0  
  repeat while (sd_response <> $00)

    ' send SD command EXIT_IDLE_STATE = $41, which resets the SD card
    sd_response := SD_Send_Command( SD_EXIT_IDLE_STATE, 0 )  

    ' test for verbose mode
    if (debug_level == DEBUG_VERBOSE)
      ' print response
      lcd.pstring(STRING("Response = "))
      lcd.hex(sd_response, 2)
      lcd.crlf
       
      lcd.pstring(STRING("Number of iterations = "))
      lcd.dec(debug_wait_res_num_iter)
      lcd.crlf
       
      lcd.pstring(STRING("Number of attempts = "))
      lcd.dec(cmd_num_attempts)
      lcd.crlf
       
    ' check if we have tried this long enough
    if (++cmd_num_attempts > SD_RESPONSE_LIMIT)
      ' test for verbose mode
      if (debug_level == DEBUG_VERBOSE)
        lcd.pstring(STRING("SD card not responding to exit idle state command."))
        lcd.crlf

      if (debug_level == DEBUG_BRIEF)
        lcd.pstring(STRING("Couldn't mount SD card!"))
        lcd.crlf
    
      return ($FF)

  ' test for verbose mode
  if (debug_level == DEBUG_VERBOSE)
    lcd.pstring(STRING("Verify SD card mode (not a MMC)..."))
    lcd.crlf
     
  ' send SD command APP_CMD = $40 + $37
  sd_response := SD_Send_Command( SD_APP_CMD, 0 )

  if ( sd_response <> $00 )
    if (debug_level == DEBUG_BRIEF)
      lcd.pstring(STRING("Couldn't mount SD card!"))
      lcd.crlf
      
    return($FF)

  ' send SD command SEND_OP_CMD = $40 + $29
  sd_response := SD_Send_Command( SD_SEND_OP_CMD, 0 )

  if ( sd_response <> $00 )
    if (debug_level == DEBUG_BRIEF)
      lcd.pstring(STRING("Couldn't mount SD card!"))
      lcd.crlf

    return($FF)

  if (debug_level == DEBUG_BRIEF)
    lcd.pstring(STRING("SD card successfully mounted."))
    lcd.crlf

  ' return success
  return($00)
  key.waitforkey
' end SD_Mount

PUB ToUpper(char)

' DESCRIPTION: this function upcases the sent character, if its not lowercase character then it passes it thru
' INPUTS: char - the character to upcase
' OUTPUTS: the upcased character
' -----------------------------------------------------------------------------
 
  if ( (char => 97) and (char =< 122))
    char -= 32
  return (char)

PUB Strncomp(string_ptr1, string_ptr2, length) | index, char

' DESCRIPTION: this function compares the two sent strings up to the nth character
' the compare is case insensitive
' INPUTS: string_ptr1 - 1st NULL terminated string to compare
'         string_ptr2 - 2ne NULL terminated string to compare 
' OUTPUTS: returns 1 if strings equal, 0 if not (case insensitive)
' -----------------------------------------------------------------------------
  
  repeat index from 0 to length-1
    ' characters are the same?
    if ( ToUpper(byte[string_ptr1][index]) <> ToUpper(byte[string_ptr2][index]) )
      ' strings differ, return failure
      return( 0 )

  ' strings must be equal      
  return( 1 )
' end Strncomp 

PUB itoa(value, sptr)  | i, z
' DESCRIPTION: converts the sent integer to an ASCII null terminated string
'
' INPUTS: value - postitive or negative integer to convert 
'         sptr  - storage for converted string
' 
' OUTPUTS: writes string to sptr and returns pointer to string as well 
' -----------------------------------------------------------------------------

  ' test for negative
  if (value < 0)
    ' add leading negative sign
    -value      
    byte[ sptr++ ] := "-"

  ' set up largest divisor
  i := 1_000_000_000
  z~

  ' compute next digit
  repeat 10
    if value => i
      byte[ sptr++ ] := value / i + "0"
      value //= i
      z~~
    elseif z or i == 1
      byte[ sptr++ ] := "0"
    i /= 10

  ' return string
  return (sptr)

' end itoa

PUB To_ASCII(inchar, replace_char)

' DESCRIPTION: converts input into a ASCII printable character, this macro function is
' is used to simply turn non-printable garbage to printable text, so the tv terminal doesn't
' get sent garbage.
'
' INPUTS: inchar - input character to test.
'         replace_char - if the input char isn't printable then replace with this character
'         (usually a dot or space, so you can see what changed) 
'
' OUTPUTS: returns the converted character
' -----------------------------------------------------------------------------

  ' test the character for printable range
  if (inchar < $20 or inchar > $7E)
    inchar := replace_char
    return(inchar)         

' end To_ASCII

PUB _Min(a,b)

' DESCRIPTION: returns signed minimum of a,b
' INPUTS: a,b - values to be compared
' OUTPUTS: the minimum
' -----------------------------------------------------------------------------

if (a < b)
  return a
else 
  return b

PUB _Max(a,b)

' DESCRIPTION: returns signed maximum of a,b
' INPUTS: a,b - values to be compared
' OUTPUTS: the maximum
' -----------------------------------------------------------------------------

if (a > b)
  return a
else 
  return b

PUB Set_Debug_Level ( level )

' DESCRIPTION: this function sets the driver debug level
' INPUTS: level - level to set output at DEBUG_BRIEF, DEBUG_OFF, DEBUG_VERBOSE
' OUPUTS: none

  ' set global debug level
  debug_level := level  

' end Set_Debug_Level  

DAT

str_newline byte 10,13

{{ LICENSE SECTION

                          TERMS OF USE: MIT License                                                  |

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation  the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:                                                                   |
                                                                                                                              |
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
}}
